# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
import re

import pytest

from autobisect.config import DEFAULT_CONFIG


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "decode_compressed_response": True,
    }


@pytest.fixture
def config_fixture(tmp_path):
    """A mock configuration file."""
    config_data = re.sub(
        r"(?<=storage-path: )(.+)",
        tmp_path.as_posix(),
        DEFAULT_CONFIG,
    )
    config_data = re.sub(r"(?<=persist-limit: )(.+)", "5", config_data)
    config_path = tmp_path / "autobisect.ini"
    config_path.write_text(config_data)

    return config_path



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from .args import JSArgs
from .js import _get_rev, JSEvaluator, JSEvaluatorException

__all__ = [
    "_get_rev",
    "JSArgs",
    "JSEvaluator",
    "JSEvaluatorException",
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from pathlib import Path
from platform import system
from unittest.mock import MagicMock

import pytest
from grizzly.common.frontend import Exit

from autobisect.evaluators import BrowserEvaluator, EvaluatorResult
from autobisect.evaluators.browser.browser import BrowserEvaluatorException


def not_linux():
    """Simple fixture for skipping parameters when not run on Linux."""
    return pytest.mark.skipif(not system().startswith("Linux"), reason="linux only")


def test_verify_build_status(mocker):
    """Test that verify_build returns the correct status."""
    mocker.patch(
        "autobisect.BrowserEvaluator.launch",
        side_effect=(
            EvaluatorResult.BUILD_CRASHED,
            EvaluatorResult.BUILD_PASSED,
            EvaluatorResult.BUILD_FAILED,
        ),
    )
    browser = BrowserEvaluator(Path("testcase.html"))

    assert browser.verify_build(Path("firefox")) is False
    assert browser.verify_build(Path("firefox")) is True
    assert browser.verify_build(Path("firefox")) is False


def test_evaluate_testcase_simple(mocker, tmp_path):
    """Test that evaluate_testcase fails when using a non-existent build path."""
    mocker.patch(
        "autobisect.BrowserEvaluator.launch",
        return_value=EvaluatorResult.BUILD_PASSED,
    )
    browser = BrowserEvaluator(Path("testcase.html"))
    binary_name = "firefox.exe" if system() == "Windows" else "firefox"
    (tmp_path / binary_name).touch()

    assert browser.evaluate_testcase(tmp_path) == EvaluatorResult.BUILD_PASSED


def test_evaluate_testcase_non_existent_binary(tmp_path):
    """Test that the binary path is calculated as "firefox.exe" on windows."""
    browser = BrowserEvaluator(Path("testcase.html"))

    assert browser.evaluate_testcase(tmp_path) == EvaluatorResult.BUILD_FAILED


def test_evaluate_testcase_system_windows(mocker, tmp_path):
    """Simple test of evaluate_testcase on windows."""
    evaluator = BrowserEvaluator(Path("testcase.html"))

    # Mock the system function to simulate running in Windows
    mocker.patch("autobisect.evaluators.browser.browser.system", return_value="Windows")

    # Mock the binary_path to an existing path
    binary_path = tmp_path / "firefox.exe"
    binary_path.touch()

    # Mock the verify_build method to return False
    spy = MagicMock(wraps=evaluator.verify_build, return_value=False)
    mocker.patch.object(evaluator, "verify_build", spy)

    evaluator.evaluate_testcase(tmp_path)

    # Assert that verify_build was called with the expected file path
    spy.assert_called_once_with(binary_path)


def test_launch_simple(mocker, tmp_path):
    """Test that launch returns the expected evaluator result."""
    mocker.patch(
        "grizzly.replay.ReplayManager.main",
        side_effect=(Exit.SUCCESS, Exit.FAILURE),
    )

    binary = tmp_path / "firefox"
    binary.touch()
    testcase = tmp_path / "testcase.html"
    testcase.touch()

    evaluator = BrowserEvaluator(testcase)

    assert evaluator.launch(binary, testcase) == EvaluatorResult.BUILD_CRASHED
    assert evaluator.launch(binary, testcase) == EvaluatorResult.BUILD_PASSED


def test_launch_non_existent_binary(tmp_path):
    """Test that launch fails when using a non-existent build path."""
    binary = tmp_path / "firefox"
    testcase = tmp_path / "testcase.html"
    testcase.touch()

    browser = BrowserEvaluator(testcase)

    with pytest.raises(BrowserEvaluatorException):
        browser.launch(binary, testcase)


@pytest.mark.parametrize("harness", (True, False))
@pytest.mark.parametrize(
    "headless", ("default", pytest.param("xvfb", marks=not_linux()))
)
@pytest.mark.parametrize("ignore", ("log-limit", "memory", "timeout"))
@pytest.mark.parametrize("pernosco", (False, pytest.param(True, marks=not_linux())))
@pytest.mark.parametrize("valgrind", (False, pytest.param(True, marks=not_linux())))
def test_grizzly_arg_parsing(
    mocker,
    tmp_path: Path,
    harness: bool,
    headless: bool,
    ignore: str,
    pernosco: bool,
    valgrind: bool,
):
    """Ensure that args are accepted by grizzly."""
    binary = tmp_path / "firefox"
    binary.touch()
    testcase = tmp_path / "testcase.html"
    testcase.touch()
    prefs = tmp_path / "prefs.js"
    prefs.touch()

    original_read_text = Path.read_text

    def read_text_bypass(self: Path, *args, **kwargs):
        """'rr needs /proc/sys/kernel/perf_event_paranoid <= 1, but it is 2'"""
        if self.absolute() == Path("/proc/sys/kernel/perf_event_paranoid"):
            return "1"
        else:
            return original_read_text(self, *args, **kwargs)

    mocker.patch(
        "grizzly.args.Path.read_text",
        autospec=True,
        side_effect=read_text_bypass,
    )

    evaluator = BrowserEvaluator(
        testcase,
        headless=headless,
        ignore=[ignore],
        launch_timeout=300,
        prefs=prefs,
        relaunch=1,
        use_harness=harness,
        use_valgrind=valgrind,
        logs=tmp_path,
        pernosco=pernosco,
        repeat=10,
    )
    evaluator.parse_args(binary, tmp_path, verify=False)


def test_grizzly_arg_parsing_no_pernosco_on_verify(tmp_path: Path):
    """Test that pernosco is not used when verifying build."""
    binary = tmp_path / "firefox"
    binary.touch()
    testcase = tmp_path / "testcase.html"
    testcase.touch()
    kwargs = {
        "pernosco": True,
        "logs": "/foo/bar",
        "repeat": 100,
    }
    evaluator = BrowserEvaluator(testcase, **kwargs)
    args = evaluator.parse_args(binary, tmp_path, verify=True)

    assert not args.pernosco



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import re
from configparser import NoOptionError, NoSectionError

import pytest

from autobisect import config


def test_config_init_no_config_file(tmp_path):
    """Initialize a config object with default settings"""
    config.APP_CONFIG_DIR = tmp_path
    config.APP_CONFIG_FILE = tmp_path / "autobisect.ini"
    config.APP_CACHE_DIR = tmp_path

    config.DEFAULT_CONFIG = re.sub(
        r"(?<=storage-path: )(.+)", tmp_path.as_posix(), config.DEFAULT_CONFIG
    )
    result = config.BisectionConfig()
    assert result.db_path == tmp_path / "autobisect.db"
    assert result.persist is True
    assert result.persist_limit == 31457280000
    assert result.store_path == tmp_path


def test_config_init_with_predefined_config(tmp_path):
    """Initialize a config object with predefined settings"""
    conf = tmp_path / "autobisect.ini"
    conf.write_text(re.sub(r"(?<=persist-limit: )(\d+)", "1000", config.DEFAULT_CONFIG))
    result = config.BisectionConfig(conf)
    assert result.persist_limit == 1048576000


def test_config_invalid_config_path(tmp_path):
    """Attempts to initialize a config object with a non-existent path"""
    with pytest.raises(IOError, match="Invalid configuration file specified"):
        config.BisectionConfig(tmp_path / "foobar")


def test_config_init_with_invalid_configuration(tmp_path):
    """Attempt to initialize a config object with an invalid configuration"""
    invalid = tmp_path / "invalid"
    invalid.touch()
    with pytest.raises((NoOptionError, NoSectionError)):
        config.BisectionConfig(invalid)


def test_config_init_with_non_existent_dir(tmp_path):
    """Attempt to initialize a config object with an invalid configuration"""
    dir_path = tmp_path / "non-existent"
    config.APP_CONFIG_DIR = dir_path
    config.APP_CONFIG_FILE = dir_path / "autobisect.ini"
    config.APP_CACHE_DIR = dir_path
    config.BisectionConfig()
    assert dir_path.is_dir()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from pathlib import Path
from platform import system

import pytest
from lithium.interestingness.timed_run import ExitStatus, RunData

from autobisect.evaluators.base import EvaluatorResult
from autobisect.evaluators.js import _get_rev, JSEvaluator, JSEvaluatorException


def test_get_rev_with_valid_fuzzmanagerconf(tmp_path):
    """Test that _get_rev returns the correct revision."""
    binary_path = tmp_path / "firefox.exe"
    fm_conf_path = binary_path.with_suffix(".fuzzmanagerconf")
    fm_conf_path.write_text("product_version = 20230908-3096b15a785a")

    assert _get_rev(binary_path) == "3096b15a785a"


def test_get_rev_with_missing_fuzzmanagerconf(tmp_path):
    """Test that _get_rev returns tip when it cannot find the product_version."""
    binary_path = tmp_path / "test_binary.exe"
    binary_path.with_suffix(".fuzzmanagerconf").touch()

    assert _get_rev(binary_path) == "tip"


def test_js_evaluator_init(tmp_path):
    """Test the initialization of JSEvaluator with valid arguments."""
    test = tmp_path / "testcase.js"
    evaluator = JSEvaluator(test, flags=["--x"], repeat=2, timeout=30, detect="crash")

    assert evaluator.testcase == test
    assert evaluator.flags == ["--x"]
    assert evaluator.repeat == 2
    assert evaluator.timeout == 30
    assert evaluator.detect == "crash"


def test_js_evaluator_init_diff_mode(tmp_path):
    """Test initialization of diff mode."""
    test = tmp_path / "testcase.js"
    evaluator = JSEvaluator(test, detect="diff", arg_1="a", arg_2="b")

    assert evaluator.detect == "diff"
    assert evaluator._arg_1 == "a"
    assert evaluator._arg_2 == "b"


def test_js_evaluator_init_diff_mode_invalid_args(tmp_path):
    """Test initialization of diff mode without supplying arg_1 or arg_2."""
    error_message = "Diff mode requires 'arg_1' and 'arg_2'"
    with pytest.raises(JSEvaluatorException, match=error_message):
        JSEvaluator(tmp_path / "testcase.js", detect="diff")


def test_js_evaluator_init_match_mode(tmp_path):
    """Test initialization of output mode."""
    test = tmp_path / "testcase.js"
    evaluator = JSEvaluator(test, detect="output", match="expected_output")

    assert evaluator.testcase == test
    assert evaluator.detect == "output"
    assert evaluator._match == "expected_output"


def test_js_evaluator_init_match_mode_invalid_args():
    """Test initialization of output mode without providing a match string."""
    error_message = "Match mode requires a match string"
    with pytest.raises(JSEvaluatorException, match=error_message):
        JSEvaluator(Path("/path/to/testcase.js"), detect="output")


@pytest.mark.vcr()
def test_js_evaluator_get_valid_flags(tmp_path):
    """Test extraction of js flags."""
    test = tmp_path / "testcase.js"
    evaluator = JSEvaluator(test, detect="crash")
    flags = evaluator.get_valid_flags("tip")

    assert sorted(flags) == [
        "arm-asm-nop-fill",
        "arm-hwcap",
        "arm-sim-icache-checks",
        "arm-sim-stop-at",
        "asm-pool-max-offset",
        "async-stacks-capture-debuggee-only",
        "available-memory",
        "baseline-eager",
        "baseline-warmup-threshold",
        "blinterp",
        "blinterp-eager",
        "blinterp-warmup-threshold",
        "cache-ir-stubs",
        "code-coverage",
        "cpu-count",
        "debug-sim",
        "delazification-mode",
        "differential-testing",
        "disable-bailout-loop-check",
        "disable-change-array-by-copy",
        "disable-jithints",
        "disable-oom-functions",
        "disable-property-error-message-fix",
        "disable-tosource",
        "disable-wasm-huge-memory",
        "disable-watchtower",
        "disable-weak-refs",
        "dll",
        "dump-entrained-variables",
        "emit-interpreter-entry",
        "enable-array-from-async",
        "enable-array-grouping",
        "enable-avx",
        "enable-ic-frame-pointers",
        "enable-import-assertions",
        "enable-iterator-helpers",
        "enable-new-set-methods",
        "enable-parallel-marking",
        "enable-shadow-realms",
        "enable-watchtower",
        "enable-well-formed-unicode-strings",
        "fast-warmup",
        "fuzzing-safe",
        "gc-param",
        "gc-zeal",
        "inlining-entry-threshold",
        "ion-check-range-analysis",
        "ion-eager",
        "ion-edgecase-analysis",
        "ion-extra-checks",
        "ion-gvn",
        "ion-inlining",
        "ion-instruction-reordering",
        "ion-iterator-indices",
        "ion-licm",
        "ion-limit-script-size",
        "ion-offthread-compile",
        "ion-optimize-gcbarriers",
        "ion-optimize-shapeguards",
        "ion-osr",
        "ion-parallel-compile",
        "ion-pruning",
        "ion-range-analysis",
        "ion-regalloc",
        "ion-scalar-replacement",
        "ion-shared-stubs",
        "ion-sink",
        "ion-warmup-threshold",
        "less-debug-code",
        "loong64-sim-icache-checks",
        "loong64-sim-stop-at",
        "marking-threads",
        "mips-sim-icache-checks",
        "mips-sim-stop-at",
        "module-load-path",
        "more-compartments",
        "no-asmjs",
        "no-async-stacks",
        "no-avx",
        "no-baseline",
        "no-blinterp",
        "no-cgc",
        "no-emit-interpreter-entry",
        "no-ggc",
        "no-incremental-gc",
        "no-ion",
        "no-ion-for-main-context",
        "no-jit-backend",
        "no-native-regexp",
        "no-source-pragmas",
        "no-sse3",
        "no-sse4",
        "no-sse41",
        "no-sse42",
        "no-ssse3",
        "no-threads",
        "no-wasm-",
        "not-implemented-watchfile",
        "nursery-bigints",
        "nursery-size",
        "nursery-strings",
        "only-inline-selfhosted",
        "regexp-warmup-threshold",
        "reprl",
        "riscv-debug",
        "riscv-sim-stop-at",
        "riscv-trap-to-simulator-debugger",
        "selfhosted-xdr-mode",
        "selfhosted-xdr-path",
        "shared-memory",
        "small-function-length",
        "smoosh",
        "spectre-mitigations",
        "suppress-minidump",
        "telemetry-dir",
        "test-wasm-await-tier2",
        "thread-count",
        "trace-regexp-assembler",
        "trace-regexp-interpreter",
        "trace-regexp-parser",
        "trace-regexp-peephole",
        "trace-sim",
        "trial-inlining-warmup-threshold",
        "use-fdlibm-for-sin-cos-tan",
        "wasm-",
        "wasm-compile-and-serialize",
        "wasm-compiler",
        "wasm-verbose",
    ]


@pytest.mark.vcr()
def test_js_evaluator_get_valid_flags_bad_rev(caplog, tmp_path):
    """Test that _get_valid_flags fails gracefully when using an invalid revision."""
    test = tmp_path / "testcase.js"
    evaluator = JSEvaluator(test, detect="crash")
    rev = "sample_rev"
    flags = evaluator.get_valid_flags(rev)

    # Assert that the function returns an empty list
    assert not flags

    # Assert that a warning message is logged with the expected exception message
    assert "Failed to retrieve build flags: " in caplog.text


@pytest.mark.parametrize("status", [True, False])
def test_js_evaluator_verify_build(mocker, tmp_path, status):
    """Test that verify_build calls timed_run with the expected arguments."""
    test = tmp_path / "testcase.js"
    evaluator = JSEvaluator(test, detect="crash", timeout=30)

    # Mock the timed_run function
    mock_timed_run = mocker.patch("lithium.interestingness.timed_run.timed_run")

    # Configure the mock return value for timed_run
    mock_run_data = mocker.MagicMock(RunData)
    mock_run_data.status = ExitStatus.NORMAL if status else ExitStatus.CRASH
    mock_timed_run.return_value = mock_run_data

    # Call the verify_build method
    binary = tmp_path / "js"
    flags = ["-flag1", "-flag2"]
    result = evaluator.verify_build(binary, flags)

    # Assert that the mock timed_run was called with the expected arguments
    mock_timed_run.assert_called_once_with([str(binary), *flags, "-e", '"quit()"'], 30)

    # Assert that the result is True because the mock timed_run returned NORMAL
    assert result is status


@pytest.mark.parametrize("verified", (True, False))
@pytest.mark.parametrize(
    "mode, ext_args",
    (
        ("crash", {}),
        ("diff", {"arg_1": "a", "arg_2": "b"}),
        ("hang", {}),
        ("output", {"match": "magic string"}),
    ),
)
def test_js_evaluator_evaluate_testcase(mocker, tmp_path, mode, ext_args, verified):
    """Simple test of evaluate_testcase."""
    (tmp_path / "dist" / "bin").mkdir(parents=True)
    binary = "js.exe" if system() == "Windows" else "js"
    (tmp_path / "dist" / "bin" / binary).touch()
    test = tmp_path / "testcase.js"

    # Set additional args required by each mode
    evaluator = JSEvaluator(test, detect=mode, flags=["--fuzzing-safe"], **ext_args)

    mocker.patch("autobisect.evaluators.js.js._get_rev", return_value="tip")
    mocker.patch.object(evaluator, "get_valid_flags", return_value=["fuzzing-safe"])
    mocker.patch.object(evaluator, "verify_build", return_value=verified)

    # Mock the timed_run function

    # Configure RunData return value for timed_run
    mock_run_data = mocker.MagicMock(RunData)
    mock_run_data.elapsedtime = 10

    mock_run_data.status = ExitStatus.NORMAL
    if mode == "diff":
        type(mock_run_data).return_code = mocker.PropertyMock(side_effect=[1, 2])
        patch_path = "lithium.interestingness.diff_test.timed_run"
    elif mode == "hang":
        mock_run_data.status = ExitStatus.TIMEOUT
        patch_path = "lithium.interestingness.hangs.timed_run"
    elif mode == "output":
        mock_run_data.err = b""
        mock_run_data.out = b"magic string"
        patch_path = "lithium.interestingness.outputs.timed_run"
    else:
        mock_run_data.err = b""
        mock_run_data.status = ExitStatus.CRASH
        patch_path = "lithium.interestingness.timed_run.timed_run"

    mocker.patch(patch_path, return_value=mock_run_data)

    result = evaluator.evaluate_testcase(tmp_path)
    if not verified:
        assert result == EvaluatorResult.BUILD_FAILED
    else:
        assert result == EvaluatorResult.BUILD_CRASHED


def test_js_evaluator_evaluate_testcase_invalid_binary(tmp_path):
    """Test that evaluate_testcase returns BUILD_FAILED for invalid binary paths."""
    binary = tmp_path / "dist" / "bin" / "js"
    test = tmp_path / "testcase.js"
    evaluator = JSEvaluator(test, detect="crash")

    assert evaluator.evaluate_testcase(binary) == EvaluatorResult.BUILD_FAILED



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Code originally taken from mozregression
# https://github.com/mozilla/mozregression/blob/5b986a3165a5208dd0722d6fc882b47e7fc1b627/mozregression/build_range.py

# The following pylint disable is needed until https://github.com/PyCQA/pylint/issues/1063 is released
# autobisect/builds.py:15:0: W0611: Unused Type imported from typing (unused-import)
# autobisect/builds.py:15:0: W0611: Unused Any imported from typing (unused-import)
# pylint: disable=unused-import

import copy
import logging
import random
from datetime import timedelta, datetime
from typing import List, Union, TypeVar, Generic, Type, Any, overload

LOG = logging.getLogger(__name__)

T = TypeVar("T")


class BuildRange(Generic[T]):
    """A class for storing a range of builds or build representations."""

    def __init__(self, builds: List[T]):
        """
        Instantiate a new instance.

        param builds: A list of Fetcher.BuildTask objects.
        """
        self._builds: List[T] = builds

    def __len__(self) -> int:
        return len(self._builds)

    @overload
    def __getitem__(self, expr: int) -> T:
        pass

    @overload
    def __getitem__(self, expr: slice) -> "BuildRange[T]":
        pass

    def __getitem__(self, expr: Union[int, slice]) -> Union[T, "BuildRange[T]"]:
        if isinstance(expr, slice):
            new_range = copy.copy(self)
            new_range._builds = self._builds[expr]
            return new_range

        return self._builds[expr]

    @property
    def builds(self) -> List[T]:
        """Returns the list of builds."""
        return self._builds

    @property
    def mid_point(self) -> Union[T, None]:
        """Returns the midpoint build"""
        if len(self) > 0:
            return self._builds[len(self) // 2]

        return None

    @property
    def random(self) -> Union[T, None]:
        """Returns a random build."""
        if len(self) > 0:
            return random.choice(self._builds)

        return None

    def index(self, build: T) -> Union[int, None]:
        """
        Returns the index of the provided build.

        :param build: An object within the "builds" list.
        :returns: The index of the provided build.
        """
        return self.builds.index(build)

    @classmethod
    def new(
        cls: "Type[BuildRange[Any]]", start: datetime, end: datetime
    ) -> "BuildRange[str]":
        """
        Creates a list of builds between two ranges
        :param start: A starting datetime object
        :type start: datetime.datetime
        :param end: An ending datetime object
        :type end: datetime.datetime
        :returns: A BuildRange object
        """
        dates = []
        # Remove time date
        start = start.replace(hour=0, minute=0, second=0, microsecond=0)
        end = end.replace(hour=0, minute=0, second=0, microsecond=0)
        delta = end - start
        for offset in range(delta.days + 1):
            dates.append((start + timedelta(days=offset)).strftime("%Y-%m-%d"))

        return cls(dates)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import logging
import os
import re
from pathlib import Path
from platform import system
from string import Template
from typing import List, Any

import requests
from lithium.interestingness import diff_test, hangs, outputs, timed_run
from lithium.interestingness.timed_run import ExitStatus

from autobisect.evaluators.base import Evaluator, EvaluatorResult

LOG = logging.getLogger(__name__)

HTTP_SESSION = requests.Session()

FLAGS_URL = Template(
    "https://hg.mozilla.org/mozilla-unified/raw-file/$rev/js/src/shell/js.cpp"
)


def _get_rev(binary: Path) -> str:
    """
    Return either the revision specified in the fuzzmanagerconf or tip.

    :param binary: Path to build.
    :returns: Extracted revision.
    """
    rev = "tip"
    path = binary.with_suffix(".fuzzmanagerconf")
    if path.is_file():
        for line in path.read_text().splitlines():
            if line.startswith("product_version"):
                rev = line.split("-")[1].strip()

    return rev


class JSEvaluatorException(Exception):
    """Raised for any JSEvaluator exception."""


class JSEvaluator(Evaluator):
    """Testcase evaluator for SpiderMonkey shells."""

    target = "js"

    def __init__(self, testcase: Path, **kwargs: Any) -> None:
        self.testcase = testcase
        self.flags = kwargs.get("flags", [])
        self.repeat = kwargs.get("repeat", 1)

        # JS Shell launch arguments
        self.timeout = kwargs.get("timeout", 60)
        self.detect = kwargs.get("detect", "crash")

        if self.detect == "diff":
            if not kwargs.get("arg_1") or not kwargs.get("arg_2"):
                raise JSEvaluatorException("Diff mode requires 'arg_1' and 'arg_2'")
            self._arg_1 = str(kwargs["arg_1"])
            self._arg_2 = str(kwargs["arg_2"])
        elif self.detect == "output":
            if not kwargs.get("match"):
                raise JSEvaluatorException("Match mode requires a match string")
            self._match = str(kwargs["match"])

    @staticmethod
    def get_valid_flags(rev: str) -> List[str]:
        """
        Extract list of runtime flags available to the current build.

        :param rev: Revision of the build.
        :returns: List of valid flags.
        """
        # Fuzzing safe is always included but not in the same format as the rest
        flags = []

        try:
            data = HTTP_SESSION.get(FLAGS_URL.substitute(rev=rev), stream=True)
            data.raise_for_status()
            matches = re.findall(r"(?:get\w+Option)\(\"(.[^\"]*)", data.text)
            flags.extend(list(set(matches)))
        except requests.exceptions.RequestException as e:
            LOG.warning("Failed to retrieve build flags: %s", e)

        return flags

    def verify_build(self, binary: Path, flags: List[str]) -> bool:
        """
        Verify that build doesn't crash on start.

        :param binary: The path to the target binary
        :param flags: Runtime flags
        :returns: True if the build is valid, False otherwise.
        """
        LOG.info("> Verifying build...")
        args = [str(binary), *flags, "-e", '"quit()"']
        run_data = timed_run.timed_run(args, self.timeout)
        if run_data.status is not ExitStatus.NORMAL:
            LOG.error(">> Build crashed!")
            return False

        return True

    def evaluate_testcase(self, build_path: Path) -> EvaluatorResult:
        """Validate build and launch with supplied testcase."""
        binary = "js.exe" if system() == "Windows" else "js"
        binary_path = build_path / "dist" / "bin" / binary
        if not binary_path.is_file():
            return EvaluatorResult.BUILD_FAILED
        rev = _get_rev(binary_path)
        all_flags = self.get_valid_flags(rev)
        flags = []
        for flag in self.flags:
            if flag.lstrip("--").split("=")[0] in all_flags:
                flags.append(flag)

        if self.verify_build(binary_path, flags):
            common_args = [
                str(binary_path),
                *flags,
                str(self.testcase),
            ]

            # Some testcases require setting the cwd to the parent dir
            previous_path = os.getcwd()
            os.chdir(os.path.dirname(os.path.abspath(self.testcase)))

            # Set LD_LIBRARY_PATH to the build directory
            previous_ld_path = os.environ.get("LD_LIBRARY_PATH", "")
            os.environ["LD_LIBRARY_PATH"] = str(binary_path.parent)

            try:
                for _ in range(self.repeat):
                    LOG.info("> Launching build with testcase...")
                    if self.detect == "diff":
                        args = ["-a", self._arg_1, "-b", self._arg_2] + common_args
                        if diff_test.interesting(args):
                            return EvaluatorResult.BUILD_CRASHED
                    elif self.detect == "output":
                        args = ["-s", self._match] + common_args
                        if outputs.interesting(args):
                            return EvaluatorResult.BUILD_CRASHED
                    elif self.detect == "crash":
                        args = [str(binary_path), *flags, str(self.testcase)]
                        result = timed_run.timed_run(args, self.timeout)
                        if result.status == ExitStatus.CRASH:
                            if b"[unhandlable oom]" not in result.err:
                                return EvaluatorResult.BUILD_CRASHED
                    elif self.detect == "hang":
                        if hangs.interesting(common_args):
                            return EvaluatorResult.BUILD_CRASHED
            finally:
                # Reset cwd and LD_LIBRARY_PATH
                os.chdir(previous_path)
                os.environ["LD_LIBRARY_PATH"] = previous_ld_path

            LOG.info("> Failed to reproduce issue!")
            return EvaluatorResult.BUILD_PASSED

        return EvaluatorResult.BUILD_FAILED



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import os
from os import access
from platform import system
from argparse import ArgumentParser, Namespace
from pathlib import Path

from autobisect.args import BisectCommonArgs


class BrowserArgs(BisectCommonArgs):
    """Arguments for the BrowserEvaluator."""

    IGNORABLE = ("log-limit", "memory", "timeout")

    def __init__(self, parser: ArgumentParser) -> None:
        super().__init__(parser)
        self.parser = parser
        launcher_grp = self.parser.add_argument_group("Launcher Arguments")
        launcher_grp.add_argument(
            "--launch-timeout",
            type=int,
            default=300,
            help="Number of seconds to wait before LaunchError is raised"
            " (default: %(default)s)",
        )
        launcher_grp.add_argument(
            "-p",
            "--prefs",
            type=Path,
            help="Optional prefs.js file to use",
        )
        display_modes = ["default", "headless"]
        if system().startswith("Linux"):
            display_modes.append("xvfb")
        launcher_grp.add_argument(
            "--display",
            choices=display_modes,
            default="default",
            help="Display mode.",
        )
        launcher_grp.add_argument(
            "--no-harness",
            action="store_true",
            help="Don't use the harness for redirection.  Browser will relaunch on every attempt.",
        )

        reporter_grp = self.parser.add_argument_group("Reporter Arguments")
        reporter_grp.add_argument(
            "--ignore",
            nargs="*",
            default=list(BrowserArgs.IGNORABLE),
            help="Space separated list of issue types to ignore. Valid options: "
            f"{' '.join(BrowserArgs.IGNORABLE)} (default: {' '.join(BrowserArgs.IGNORABLE)})",
        )

        if system().startswith("Linux"):
            dbg_group = launcher_grp.add_mutually_exclusive_group()
            dbg_group.add_argument(
                "--pernosco",
                action="store_true",
                help="Use rr. Trace intended to be used with Pernosco.",
            )
            dbg_group.add_argument("--rr", action="store_true", help="Use rr.")
        else:
            self.parser.set_defaults(
                pernosco=False,
                rr=False,
                valgrind=False,
            )

    def sanity_check(self, args: Namespace) -> None:
        """
        Perform Sanity Checks.

        :param args: Parsed arguments.
        """
        super().sanity_check(args)
        if args.prefs:
            args.prefs = args.prefs.expanduser()
            if not (args.prefs.is_file() and access(args.prefs, os.R_OK)):
                self.parser.error("Cannot read the prefs.js file!")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import logging
import time
from argparse import ArgumentParser, Namespace
from datetime import timedelta
from typing import Union, Optional, List

from fuzzfetch import BuildFlags, Platform
from grizzly.main import configure_logging

from autobisect.bisect import BisectionResult, Bisector
from autobisect.evaluators import BrowserArgs, BrowserEvaluator, JSArgs, JSEvaluator

LOG = logging.getLogger("autobisect")


def parse_args(argv: Optional[List[str]] = None) -> Namespace:
    """Argument parser."""
    parser = ArgumentParser(description="Firefox and Spidermonkey Bisection Tool")
    subparsers = parser.add_subparsers(dest="target", required=True)
    firefox_parser = BrowserArgs(subparsers.add_parser("firefox"))
    js_parser = JSArgs(subparsers.add_parser("js"))

    args = parser.parse_args(argv)
    if args.target == "firefox":
        firefox_parser.sanity_check(args)
    elif args.target == "js":
        js_parser.sanity_check(args)

    return args


def main(argv: Optional[List[str]] = None) -> int:
    """
    Main entry point.

    :param argv: Parsed arguments.
    :returns: Exit code.
    """
    args = parse_args(argv)
    configure_logging(args.log_level)

    evaluator: Union[BrowserEvaluator, JSEvaluator]
    if args.target == "firefox":
        evaluator = BrowserEvaluator(**vars(args))
    else:
        args.flags = [] if args.flags is None else args.flags.split(" ")
        evaluator = JSEvaluator(**vars(args))

    flags = BuildFlags(
        args.asan,
        args.tsan,
        args.debug,
        args.fuzzing,
        args.coverage,
        args.valgrind,
        args.no_opt,
        args.fuzzilli,
    )
    platform = Platform(args.os, args.cpu)
    bisector = Bisector(
        evaluator,
        args.branch,
        args.start,
        args.end,
        flags,
        platform,
        args.find_fix,
        args.config,
    )
    start_time = time.time()
    result = bisector.bisect()
    end_time = time.time()
    if result.status == BisectionResult.SUCCESS:
        LOG.info("Reduced build range to:")
        LOG.info("> Start: %s (%s)", result.start.changeset, result.start.id)
        LOG.info("> End: %s (%s)", result.end.changeset, result.end.id)
        LOG.info("> %s", result.pushlog)
    else:
        LOG.error("Bisection failed!")

    elapsed = timedelta(seconds=int(end_time - start_time))
    LOG.info("Bisection completed in: %s", elapsed)

    return 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
import platform

import pytest

from autobisect.main import parse_args


@pytest.mark.parametrize("target", ("firefox", "js"))
def test_parse_args_simple(tmp_path, target):
    """Test that parse_args works with minimal arguments"""
    testcase = tmp_path / "testcase"
    testcase.touch()
    args = parse_args([target, str(testcase)])
    assert args.target == target


@pytest.mark.parametrize("target", ("firefox", "js"))
def test_parse_args_invalid_testcase_path(tmp_path, target):
    """Test that parse_args raises when using an invalid testcase path"""
    with pytest.raises(SystemExit):
        parse_args([target, str(tmp_path / "non-existent")])


def test_parse_args_invalid_target():
    """Test that parse_args raises when using an invalid target"""
    with pytest.raises(SystemExit):
        parse_args(["foo"])


@pytest.mark.parametrize(
    "args",
    [
        (["firefox", "--log-level", "INVALID"], "Invalid log-level"),
        (["firefox", "--config", "404.ini"], "Cannot access configuration file!"),
        (["firefox", "--fuzzilli"], "Fuzzilli builds are not available for firefox"),
    ],
)
def test_parse_args_raises_with_invalid_args(capsys, tmp_path, args):
    """Test that parse_args raises when using invalid arguments"""
    testcase = tmp_path / "testcase"
    testcase.touch()
    args[0].append(str(testcase))
    with pytest.raises(SystemExit):
        parse_args(args[0])

    _, err = capsys.readouterr()
    assert args[1] in err


@pytest.mark.skipif(platform.system() != "Windows", reason="Windows only")
def test_parse_args_windows_defaults(mocker, tmp_path):
    """Test that parse_args sets the expected windows defaults"""
    testcase = tmp_path / "testcase.html"
    testcase.touch()

    args = parse_args(["firefox", str(testcase)])
    assert not args.pernosco
    assert not args.rr
    assert not args.valgrind


def test_parse_args_prefs_sanity_check(mocker, tmp_path):
    """Test that parse_args exits if the prefs.js file is not accessible."""
    testcase = tmp_path / "testcase.html"
    testcase.touch()
    prefs = tmp_path / "prefs.js"

    with pytest.raises(SystemExit):
        parse_args(["firefox", str(testcase), "--prefs", str(prefs)])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Generator, Optional, List, Union, TypeVar, Callable

import requests
from fuzzfetch import (
    BuildFlags,
    BuildSearchOrder,
    BuildTask,
    Fetcher,
    FetcherException,
    Platform,
)

from .build_manager import BuildManager, BuildManagerException
from .builds import BuildRange
from .evaluators import Evaluator, EvaluatorResult

T = TypeVar("T")

LOG = logging.getLogger(__name__)


def get_autoland_range(start: str, end: str) -> Union[List[str], None]:
    """
    Retrieve changeset from autoland within supplied boundary.

    :param start: Starting revision.
    :param end: Ending revision.
    :return: List of changesets.
    """
    url = (
        "https://hg.mozilla.org/mozilla-central/json-pushes"
        f"?fromchange={start}&tochange={end}"
    )
    try:
        data = requests.get(url, timeout=30)
        data.raise_for_status()
    except requests.exceptions.RequestException as exc:
        LOG.error("Failed to retrieve autoland changeset %s", exc)
        return None

    json = data.json()

    changesets = []
    for push_id in json.keys():
        changesets.extend(json[push_id]["changesets"])

    return changesets


class StatusException(Exception):
    """Raised when an invalid status is supplied."""


class VerificationStatus(Enum):
    """Class for storing build verification result."""

    SUCCESS = 0
    START_BUILD_FAILED = 1
    END_BUILD_FAILED = 2
    START_BUILD_CRASHES = 3
    END_BUILD_PASSES = 4
    FIND_FIX_START_BUILD_PASSES = 5
    FIND_FIX_END_BUILD_CRASHES = 6

    @property
    def message(self) -> Optional[str]:
        """Return message matching explaining current status."""
        result = None
        if self == VerificationStatus.SUCCESS:
            result = "Verified supplied boundaries!"
        elif self == VerificationStatus.START_BUILD_FAILED:
            result = "Unable to launch the start build!"
        elif self == VerificationStatus.END_BUILD_FAILED:
            result = "Unable to launch the end build!"
        elif self == VerificationStatus.START_BUILD_CRASHES:
            result = "Testcase reproduces on start build!"
        elif self == VerificationStatus.END_BUILD_PASSES:
            result = "Testcase does not reproduce on end build!"
        elif self == VerificationStatus.FIND_FIX_START_BUILD_PASSES:
            result = "Start build didn't crash!"
        elif self == VerificationStatus.FIND_FIX_END_BUILD_CRASHES:
            result = "End build crashes!"

        return result


class BisectionResult(object):
    """Class for storing bisection result."""

    SUCCESS = 0
    FAILED = 1

    def __init__(
        self,
        status: int,
        start: Fetcher,
        end: Fetcher,
        branch: str,
        message: Optional[str] = None,
    ):
        self.status = status
        self.start = start
        self.end = end
        self.branch = branch
        if status == BisectionResult.SUCCESS:
            base = "https://hg.mozilla.org/mozilla-unified"
            if start.build_info["moz_source_repo"] == end.build_info["moz_source_repo"]:
                base = start.build_info["moz_source_repo"]
            elif self.branch == "central" and "autoland" in (
                start._branch,
                end._branch,
            ):
                for build in (start, end):
                    if build._branch == "central":
                        base = build.build_info["moz_source_repo"]
                        break

            self.pushlog = (
                f"{base}/pushloghtml?fromchange="
                f"{start.changeset}&tochange={end.changeset}"
            )

        self.message = message


class Bisector(object):
    """Taskcluster Bisection Class."""

    def __init__(
        self,
        evaluator: Evaluator,
        branch: str,
        start: Union[str, None],
        end: Union[str, None],
        flags: BuildFlags,
        platform: Platform,
        find_fix: bool = False,
        config: Optional[Path] = None,
    ):
        """
        Instantiate bisection object.

        :param evaluator: Object instance used to evaluate testcase.
        :param branch: Mozilla branch to use for finding builds.
        :param start: Start revision, date, or buildid.
        :param end: End revision, date, or buildid.
        :param flags: Build flags (asan, tsan, debug, fuzzing, valgrind).
        :param platform: fuzzfetch.fetch.Platform instance.
        :param find_fix: Boolean identifying whether to find a fix or bisect bug.
        :param config: Path to config file.
        """
        self.evaluator: Evaluator = evaluator
        self.branch = branch
        self.platform: Platform = platform
        self.flags = flags
        self.find_fix = find_fix

        # If no start date is supplied, default to the oldest available build
        max_days = 364 if evaluator.target == "firefox" else 89
        earliest = (datetime.utcnow() - timedelta(days=max_days)).strftime("%Y-%m-%d")

        start_id = start if start else earliest
        end_id = end if end else "latest"

        self.start = Fetcher(
            self.branch,
            start_id,
            self.flags,
            targets=[self.evaluator.target],
            platform=self.platform,
            nearest=BuildSearchOrder.ASC,
        )
        self.end = Fetcher(
            self.branch,
            end_id,
            self.flags,
            targets=[self.evaluator.target],
            platform=self.platform,
            nearest=BuildSearchOrder.DESC,
        )

        self.build_manager = BuildManager(config)

    def _get_daily_builds(self) -> BuildRange[str]:
        """Create build range containing one build per day."""
        start = self.start.datetime + timedelta(days=1)
        end = self.end.datetime - timedelta(days=1)
        LOG.info(f"Enumerating daily builds: {start} - {end}")

        return BuildRange.new(start, end)

    def _get_pushdate_builds(self) -> BuildRange[Fetcher]:
        """Create build range containing all builds per pushdate."""
        start = self.start.datetime
        end = self.end.datetime
        LOG.info(f"Enumerating pushdate builds: {start} - {end}")

        builds = []
        for dt in [start, end]:
            date = dt.strftime("%Y-%m-%d")
            for task in BuildTask.iterall(date, self.branch, self.flags, self.platform):
                # Ignore "latest" as these are aliases for the most recent build
                if hasattr(task, "url") and ".latest." in str(task.url):
                    continue

                # Only keep builds after the start and before the end boundaries
                build = Fetcher(
                    self.branch,
                    task,
                    self.flags,
                    targets=[self.evaluator.target],
                    platform=self.platform,
                )
                if self.end.datetime > build.datetime > self.start.datetime:
                    if build.changeset not in (
                        self.start.changeset,
                        self.end.changeset,
                    ):
                        builds.append(build)

        return BuildRange(builds)

    def _get_autoland_builds(self) -> BuildRange[Fetcher]:
        """Create build range containing all autoland builds per pushdate"""
        if self.branch != "central":
            return BuildRange([])

        start = self.start.datetime
        end = self.end.datetime

        LOG.info(f"Enumerating autoland builds: {start} - {end}")
        changesets = get_autoland_range(self.start.changeset, self.end.changeset)
        if changesets is None:
            return BuildRange([])

        builds = []
        for changeset in changesets:
            try:
                build = Fetcher(
                    "autoland",
                    changeset,
                    self.flags,
                    targets=[self.evaluator.target],
                    platform=self.platform,
                )
                builds.append(build)
            except FetcherException:
                LOG.warning("Unable to find build for %s", changeset)

        return BuildRange(builds)

    def build_iterator(
        self,
        build_range: BuildRange[Union[str, Fetcher]],
        random_choice: bool,
    ) -> Generator[Fetcher, EvaluatorResult, None]:
        """Yields next build to be evaluated until all possibilities consumed."""
        while build_range:
            if random_choice:
                build = build_range.random
            else:
                build = build_range.mid_point

            assert build is not None
            index = build_range.index(build)
            if not isinstance(build, Fetcher):
                try:
                    build = Fetcher(
                        self.branch,
                        build,
                        self.flags,
                        targets=[self.evaluator.target],
                        platform=self.platform,
                    )
                except FetcherException:
                    LOG.warning("Unable to find build for %s", build)
                    build_range.builds.remove(build)
                    continue

            status = yield build

            assert isinstance(index, int)
            build_range = self.update_range(status, build, index, build_range)

    def bisect(self, random_choice: bool = False) -> BisectionResult:
        """
        Main bisection function.

        :param random_choice: Select builds at random during bisection (QuickSort).
        :returns: Bisection result.
        """
        LOG.info("Begin bisection...")
        LOG.info("> Start: %s (%s)", self.start.changeset, self.start.id)
        LOG.info("> End: %s (%s)", self.end.changeset, self.end.id)

        verified = self.verify_bounds()
        if verified == VerificationStatus.SUCCESS:
            LOG.info(verified.message)
        else:
            LOG.critical(verified.message)
            return BisectionResult(
                BisectionResult.FAILED,
                self.start,
                self.end,
                self.branch,
                verified.message,
            )

        LOG.info("Attempting to reduce bisection range using taskcluster binaries")
        strategies: List[Callable[[], Union[BuildRange[str], BuildRange[Fetcher]]]] = [
            self._get_daily_builds,
            self._get_pushdate_builds,
        ]
        if self.branch == "central":
            strategies.append(self._get_autoland_builds)

        for strategy in strategies:
            build_range = strategy()
            generator = self.build_iterator(build_range, random_choice)  # type: ignore
            try:
                next_build = next(generator)
                while True:
                    status = self.test_build(next_build)
                    next_build = generator.send(status)
            except StopIteration:
                pass

        return BisectionResult(
            BisectionResult.SUCCESS, self.start, self.end, self.branch
        )

    def update_range(
        self,
        status: EvaluatorResult,
        build: Fetcher,
        index: int,
        build_range: BuildRange[T],
    ) -> BuildRange[T]:
        """
        Returns a new build range based on the status of the previously evaluated test.

        :param status: The status of the evaluated testcase.
        :param build: The evaluated build.
        :param index: Index of the build.
        :param build_range: The build_range to update.
        :returns: New build range.
        """
        if status == EvaluatorResult.BUILD_PASSED:
            if not self.find_fix:
                self.start = build
                return build_range[index + 1 :]

            self.end = build
            return build_range[:index]
        if status == EvaluatorResult.BUILD_CRASHED:
            if not self.find_fix:
                self.end = build
                return build_range[:index]

            self.start = build
            return build_range[index + 1 :]
        if status == EvaluatorResult.BUILD_FAILED:
            range_copy = build_range[:]
            range_copy.builds.pop(index)
            return range_copy

        raise StatusException("Invalid status supplied")

    def test_build(self, build: Fetcher) -> EvaluatorResult:
        """
        Prepare the build directory and launch the supplied build.

        :param build: A Fetcher object to prevent duplicate fetching
        :return: The result of the build evaluation
        """
        LOG.info("Testing build %s (%s)", build.changeset, build.id)
        # If persistence is enabled and a build exists, use it
        try:
            with self.build_manager.get_build(build, self.evaluator.target) as path:
                return self.evaluator.evaluate_testcase(path)
        except BuildManagerException:
            return EvaluatorResult.BUILD_FAILED

    def verify_bounds(self) -> VerificationStatus:
        """Verify that the supplied bounds behave as expected"""
        LOG.info("Attempting to verify boundaries...")
        start_result = self.test_build(self.start)
        if start_result not in set(EvaluatorResult):
            raise StatusException("Invalid status supplied")

        if start_result == EvaluatorResult.BUILD_FAILED:
            return VerificationStatus.START_BUILD_FAILED
        if start_result == EvaluatorResult.BUILD_CRASHED and not self.find_fix:
            return VerificationStatus.START_BUILD_CRASHES
        if start_result == EvaluatorResult.BUILD_PASSED and self.find_fix:
            return VerificationStatus.FIND_FIX_START_BUILD_PASSES

        end_result = self.test_build(self.end)
        if end_result not in set(EvaluatorResult):
            raise StatusException("Invalid status supplied")

        if end_result == EvaluatorResult.BUILD_FAILED:
            return VerificationStatus.END_BUILD_FAILED
        if end_result == EvaluatorResult.BUILD_PASSED and not self.find_fix:
            return VerificationStatus.END_BUILD_PASSES
        if end_result == EvaluatorResult.BUILD_CRASHED and self.find_fix:
            return VerificationStatus.FIND_FIX_END_BUILD_CRASHES

        return VerificationStatus.SUCCESS



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path


class EvaluatorResult(Enum):
    """Evaluator result."""

    BUILD_CRASHED = 0
    BUILD_PASSED = 1
    BUILD_FAILED = 2


class Evaluator(ABC):
    """Base evaluator class."""

    @property
    @abstractmethod
    def target(self) -> str:
        """The corresponding Fetcher target."""

    @abstractmethod
    def evaluate_testcase(self, build_path: Path) -> EvaluatorResult:
        """Method for evaluating testcase."""



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import logging
import os
import shutil
import sqlite3
import time
from contextlib import contextmanager
from pathlib import Path
from typing import List, Optional, Iterator

from fuzzfetch import Fetcher

from autobisect.config import BisectionConfig

LOG = logging.getLogger(__name__)


class BuildManagerException(Exception):
    """Raised when a build cannot be retrieved."""


class DatabaseManager(object):
    """Sqlite3 wrapper class."""

    def __init__(self, db_path: Path) -> None:
        self.con = sqlite3.connect(str(db_path))
        self.cur = self.con.cursor()
        self.cur.execute("CREATE TABLE IF NOT EXISTS in_use (build_path TEXT, pid INT)")
        self.cur.execute(
            "CREATE TABLE IF NOT EXISTS download_queue (build_path TEXT primary key, pid INT)"
        )

    def close(self) -> None:
        """Closes the sqlite3 database."""
        if self.con:
            self.con.commit()
            self.con.close()

    def __del__(self) -> None:
        try:
            self.close()
        except sqlite3.ProgrammingError:
            pass


class BuildManager(object):
    """A class for managing downloaded builds."""

    def __init__(self, config: Optional[Path] = None) -> None:
        self.config = BisectionConfig(config)
        self.build_dir = self.config.store_path / "builds"
        if not Path.is_dir(self.build_dir):
            self.build_dir.mkdir(parents=True)

        self.pid = os.getpid()
        self.db = DatabaseManager(self.config.db_path)

    @property
    def current_build_size(self) -> int:
        """Return the total size of all cached builds."""
        return sum(os.path.getsize(f) for f in self.build_dir.rglob("*"))

    def enumerate_builds(self) -> List[Path]:
        """
        Enumerate all available builds.

        :returns: A list of all available builds sorted by access time.
        """
        builds = [x for x in self.build_dir.iterdir() if x.is_dir()]
        return sorted(builds, key=lambda b: b.stat().st_atime_ns)

    def remove_old_builds(self) -> None:
        """Removes stored builds to make room for newer builds."""
        while self.current_build_size > self.config.persist_limit:
            builds = self.enumerate_builds()
            for build_path in builds:
                if self.current_build_size < self.config.persist_limit:
                    break
                res = self.db.cur.execute(
                    "SELECT COUNT(1) FROM ("
                    "SELECT build_path FROM in_use UNION ALL "
                    "SELECT build_path FROM download_queue"
                    ") WHERE build_path == ?",
                    (str(build_path),),
                )
                result = res.fetchone()
                assert result is not None
                build_in_use = result[0]
                if not build_in_use:
                    LOG.debug("Removing build: %s", build_path)
                    shutil.rmtree(build_path)
                self.db.con.commit()

            time.sleep(0.1)

    @contextmanager
    def get_build(self, build: Fetcher, target: str) -> Iterator[Path]:
        """
        Retrieve the build matching the supplied revision.

        :param build: A fuzzFetch.Fetcher build object.
        :param target: The target to retrieve (i.e. firefox, js, gtest, etc.).
        :yields: The build path.
        """
        # pylint: disable=protected-access
        branch = f"m-{build._branch[0]}"
        platform = build._platform.system
        flags = build._flags.build_string()
        rev = build.changeset[:12]
        build_name = f"{target}-{branch}-{platform}{flags}-{rev}"
        target_path = self.build_dir / build_name.lower()
        path_string = os.fspath(target_path)

        try:
            # Insert build_path into in_use to prevent deletion
            self.db.cur.execute(
                "INSERT INTO in_use VALUES (?, ?)", (path_string, self.pid)
            )
            self.db.con.commit()

            # Try to insert the build_path into download_queue
            # If the insert fails, another process is already downloading it
            # Poll the database until it complete
            try:
                self.db.cur.execute(
                    "INSERT into download_queue VALUES (?, ?)", (path_string, self.pid)
                )
                self.db.con.commit()
                # If the build doesn't exist on disk, download it
                if not Path.is_dir(target_path):
                    self.remove_old_builds()
                    build.extract_build(target_path)
            except sqlite3.IntegrityError:
                LOG.warning(
                    "Another process is attempting to download the build. Waiting"
                )
                while True:
                    res = self.db.cur.execute(
                        "SELECT * FROM download_queue WHERE build_path = ?",
                        (path_string,),
                    )
                    if res.fetchone() is None:
                        break

                    time.sleep(0.1)
            finally:
                self.db.cur.execute(
                    "DELETE FROM download_queue WHERE build_path = ? AND pid = ?",
                    (path_string, self.pid),
                )
                self.db.con.commit()

            yield target_path
        finally:
            self.db.cur.execute(
                "DELETE FROM in_use WHERE build_path = ? AND pid = ?",
                (path_string, self.pid),
            )
            self.db.con.commit()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import sqlite3
from pathlib import Path

import pytest
from fuzzfetch import Fetcher, BuildFlags

from autobisect.build_manager import (
    BuildManager,
    DatabaseManager,
)


def test_database_manager_init_new_db(tmp_path):
    """Initialize new database instance"""
    db_path = tmp_path / "foo.db"
    db = DatabaseManager(db_path)
    assert db_path.is_file()
    assert isinstance(db.con, sqlite3.Connection)
    assert isinstance(db.cur, sqlite3.Cursor)


def test_database_manager_close(tmp_path):
    """Test that the connection is closed"""
    db = DatabaseManager(tmp_path / "foo.db")
    db.close()
    with pytest.raises(sqlite3.ProgrammingError) as e:
        db.cur.execute("SELECT * FROM FOO")

    assert str(e.value) == "Cannot operate on a closed database."


def test_database_manager_destructor(tmp_path):
    """Test that the database cursor is cleaned up on destruction"""
    db = DatabaseManager(tmp_path / "foo.db")
    ref = db.con
    del db

    with pytest.raises(sqlite3.ProgrammingError) as e:
        ref.execute("SELECT * FROM FOO")

    assert str(e.value) == "Cannot operate on a closed database."


def test_build_manager_init():
    """Initialize new BuildManager instance"""
    manager = BuildManager()
    assert manager.build_dir.is_dir()
    assert isinstance(manager.pid, int)
    assert isinstance(manager.db, DatabaseManager)


def test_build_manager_init_with_config(config_fixture):
    """Initialize new BuildManager instance with a config"""
    manager = BuildManager(config_fixture)
    assert manager.build_dir.is_dir()


def test_build_manager_build_size(mocker, config_fixture):
    """Simple test of BuildManager.build_size"""
    # Patch getsize to return a static value of 1MB
    mocker.patch("os.path.getsize", return_value=1024 * 1024)

    manager = BuildManager(config_fixture)
    total = 10
    for i in range(total):
        build_dir = Path(manager.build_dir / str(i))
        build_dir.mkdir()
    assert manager.current_build_size == 1024 * 1024 * total


def test_build_manager_enumerate_builds(config_fixture):
    """Simple test of BuildManager.enumerate_builds"""
    manager = BuildManager(config_fixture)

    assert not manager.enumerate_builds()

    builds = []
    total = 10
    for i in range(total):
        build_dir = manager.build_dir / f"firefox_{i}"
        build_dir.mkdir()
        builds.append(build_dir)

    assert len(manager.enumerate_builds()) == total


def test_build_manager_remove_old_builds(mocker, config_fixture):
    """Simple test of BuildManager.remove_old_builds"""
    # Patch getsize to return a static value of 1MB
    mocker.patch("os.path.getsize", return_value=1024 * 1024)

    # Create 10 mock builds for a total of 10MBs
    manager = BuildManager(config_fixture)
    total = 10
    for i in range(total):
        build_dir = manager.build_dir / f"firefox_{i}"
        build_dir.mkdir()

    manager.remove_old_builds()
    # With a storage limit of 5MBs,
    # the result should be one less than half
    expected = total / 2 - 1
    assert len(manager.enumerate_builds()) == expected
    assert manager.current_build_size == 1024 * 1024 * expected


def test_build_manager_remove_old_builds_in_use(mocker, config_fixture):
    """Test that in_use builds are not removed"""
    # Patch getsize to return a static value of 1MB
    mocker.patch("os.path.getsize", return_value=1024 * 1024)

    manager = BuildManager(config_fixture)
    total = 6
    for i in range(total):
        build_dir = manager.build_dir / f"firefox_{i}"
        build_dir.mkdir()
        if i != 0:
            # Mark all builds other than the first as in_use
            manager.db.cur.execute(
                "INSERT INTO in_use VALUES (?, ?)", (str(build_dir), 1)
            )
            manager.db.con.commit()

    manager.remove_old_builds()
    assert manager.build_dir / "firefox_0" not in manager.enumerate_builds()


@pytest.mark.vcr()
def test_build_manager_get_build(mocker, config_fixture):
    """Simple test of BuildManager.get_build"""
    manager = BuildManager(config_fixture)
    flags = BuildFlags()
    fetcher = Fetcher("central", "latest", flags, ["firefox"])
    extract_build = mocker.patch.object(Fetcher, "extract_build")

    execute = manager.db.cur.execute
    with manager.get_build(fetcher, "firefox") as build:
        assert build is not None
        assert extract_build.call_count == 1
        # The build should be marked as in_use
        res = execute("SELECT * FROM in_use WHERE build_path == ?", (str(build),))
        assert res.fetchone() is not None

    # The build should no longer be marked as in_use
    res = execute("SELECT * FROM in_use WHERE build_path == ?", (str(build),))
    assert res.fetchone() is None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
import re
from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest
from fuzzfetch import BuildFlags, Fetcher, Platform

from autobisect import EvaluatorResult, BrowserEvaluator
from autobisect.bisect import (
    Bisector,
    StatusException,
    VerificationStatus,
    get_autoland_range,
)
from autobisect.builds import BuildRange


class MockFetcher:
    """Class for mocking Fetcher objects."""

    def __init__(self, dt=None, changeset=None):
        self.datetime = dt
        self.changeset = changeset


class MockBisector(Bisector):
    """Class for mocking Bisector objects."""

    # pylint: disable=super-init-not-called
    def __init__(self, start: datetime, end: datetime):
        self.start = MockFetcher(dt=start)
        self.end = MockFetcher(dt=end)
        self.branch = "central"
        self.find_fix = False
        self.flags = BuildFlags()
        self.platform = Platform("Linux", "x86_64")
        self.evaluator = BrowserEvaluator(Path("testcase.html"))


@pytest.mark.freeze_time("2024-05-30")
@pytest.mark.parametrize("delta, expected", [[0, 0], [11, 10]])
@pytest.mark.vcr()
def test_bisect_get_daily_builds_simple(delta, expected):
    """Test that get_daily_builds returns the expected build range."""
    start_date = datetime.now(tz=timezone.utc) - timedelta(days=delta)
    end_date = datetime.now(tz=timezone.utc)
    bisector = MockBisector(start_date, end_date)
    builds = bisector._get_daily_builds()

    assert isinstance(builds, BuildRange)
    assert all(re.match(r"\d{4}-\d{2}-\d{2}", b) is not None for b in builds)
    assert len(builds) == expected


@pytest.mark.freeze_time("2024-05-30")
@pytest.mark.vcr()
def test_bisect_get_pushdate_builds_simple():
    """Test that get_daily_builds returns the expected build range."""
    start_date = datetime.now(tz=timezone.utc) - timedelta(days=1)
    end_date = datetime.now(tz=timezone.utc)
    bisector = MockBisector(start_date, end_date)

    builds = bisector._get_pushdate_builds()

    assert len(builds) == 3
    assert isinstance(builds, BuildRange)
    for build in builds:
        assert isinstance(build, Fetcher)


@pytest.mark.freeze_time("2024-05-30")
@pytest.mark.vcr()
@pytest.mark.skip(reason="Cassette is too large")
def test_bisect_get_autoland_builds_simple(browser_evaluator, opt_flags, platform):
    """Test that get_autoland_builds returns the expected build range."""
    start = "abdb14987e23f88107f437b130d0817eea873199"
    end = "b05a24f158503b7ea175520ef383fa005a4c41b7"
    bisector = Bisector(
        browser_evaluator,
        "central",
        start,
        end,
        opt_flags,
        platform,
    )
    builds = bisector._get_autoland_builds()
    repo_url = "https://hg.mozilla.org/integration/autoland"

    assert isinstance(builds, BuildRange)
    assert len(builds) == 10
    assert all(b.build_info["moz_source_repo"] == repo_url for b in builds)


@pytest.mark.vcr()
def test_get_autoland_range_simple():
    """Test that get_autoland_range returns a list of changesets."""
    start = "abdb14987e23f88107f437b130d0817eea873199"
    end = "b05a24f158503b7ea175520ef383fa005a4c41b7"
    changesets = get_autoland_range(start, end)

    assert isinstance(changesets, list)
    assert len(changesets) == 35
    assert all(isinstance(c, str) and len(c) == 40 for c in changesets)


@pytest.mark.vcr()
def test_get_autoland_range_invalid_revs():
    """Test that get_autoland_range returns None when using invalid revisions."""
    assert get_autoland_range("foo", "bar") is None


@pytest.mark.parametrize("status", EvaluatorResult)
@pytest.mark.parametrize("find_fix", [True, False])
def test_update_range_simple(status, find_fix):
    """Test that update_range returns the correct build range based on status."""
    builds = []
    for _ in range(10):
        next_date = datetime(2020, 1, 1, 0, 0) + timedelta(days=1)
        builds.append(MockFetcher(dt=next_date))

    for index, build in enumerate(builds):
        bisector = MockBisector(builds[0].datetime, builds[-1].datetime)
        bisector.find_fix = find_fix
        build_range = BuildRange(builds)
        bisector.update_range(status, build, index, build_range)
        if status == EvaluatorResult.BUILD_PASSED:
            if find_fix:
                assert bisector.end == build
            else:
                assert bisector.start == build
        elif status == EvaluatorResult.BUILD_CRASHED:
            if find_fix:
                assert bisector.start == build
            else:
                assert bisector.end == build


@pytest.mark.parametrize("find_fix", [True, False])
@pytest.mark.parametrize("end_result", EvaluatorResult)
@pytest.mark.parametrize("start_result", EvaluatorResult)
# pylint: disable=inconsistent-return-statements
def test_verify_bounds_simple(mocker, start_result, end_result, find_fix):
    """Test that verify_bounds returns the expected status."""
    bisector = MockBisector(datetime.now(), datetime.now())
    bisector.find_fix = find_fix

    mocker.patch(
        "autobisect.bisect.Bisector.test_build", side_effect=[start_result, end_result]
    )
    result = bisector.verify_bounds()

    if start_result == EvaluatorResult.BUILD_FAILED:
        assert result == VerificationStatus.START_BUILD_FAILED
    elif start_result == EvaluatorResult.BUILD_CRASHED and not find_fix:
        assert result == VerificationStatus.START_BUILD_CRASHES
    elif start_result == EvaluatorResult.BUILD_PASSED and find_fix:
        assert result == VerificationStatus.FIND_FIX_START_BUILD_PASSES
    elif end_result == EvaluatorResult.BUILD_FAILED:
        assert result == VerificationStatus.END_BUILD_FAILED
    elif end_result == EvaluatorResult.BUILD_PASSED and not find_fix:
        assert VerificationStatus.END_BUILD_PASSES
    elif end_result == EvaluatorResult.BUILD_CRASHED and find_fix:
        assert VerificationStatus.FIND_FIX_END_BUILD_CRASHES
    else:
        assert result == VerificationStatus.SUCCESS


@pytest.mark.parametrize(
    "test_results",
    [[EvaluatorResult.BUILD_PASSED, None], [None, EvaluatorResult.BUILD_CRASHED]],
)
def test_verify_bounds_invalid_status(mocker, test_results):
    """
    Test that verify_bounds throw a StatusException when start build passes or end
    build crashes.
    """
    bisector = MockBisector(datetime.now(), datetime.now())
    mocker.patch("autobisect.bisect.Bisector.test_build", side_effect=test_results)
    with pytest.raises(StatusException):
        bisector.verify_bounds()


def test_build_iterator_random(mocker):
    """Test that builds are selected at random when random_choice arg is set."""
    builds = BuildRange([])
    for _ in range(1, 4):
        builds._builds.append(mocker.Mock(spec=Fetcher))
    spy = mocker.patch("autobisect.builds.random.choice", side_effect=builds)

    bisector = MockBisector(datetime.now(), datetime.now())
    generator = bisector.build_iterator(builds, True)
    try:
        next(generator)
        while True:
            generator.send(EvaluatorResult.BUILD_PASSED)
    except StopIteration:
        pass

    assert spy.call_count == 3


def test_verification_status_message():
    """Test that VerificationStatus always returns a message."""
    assert len(VerificationStatus) == 7
    for entry in VerificationStatus:
        assert isinstance(VerificationStatus[entry.name].message, str)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
import itertools
import os
import platform as std_platform
from argparse import ArgumentParser, Namespace
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from pathlib import Path

from fuzzfetch import Fetcher, Platform

CPU_CHOICES = sorted(
    set(
        itertools.chain(
            itertools.chain.from_iterable(Platform.SUPPORTED.values()),
            Platform.CPU_ALIASES,
        )
    )
)

LOG_LEVELS = {
    "CRIT": CRITICAL,
    "ERROR": ERROR,
    "WARN": WARNING,
    "INFO": INFO,
    "DEBUG": DEBUG,
}


class BisectCommonArgs:
    """Arguments common to all bisection targets."""

    def __init__(self, parser: ArgumentParser):
        """
        Add common args to parser.
        :param parser: Base parser.
        """
        self.parser = parser
        self.parser.add_argument(
            "testcase",
            type=Path,
            help="Path to testcase",
        )

        self.parser.add_argument(
            "--log-level",
            default="INFO",
            help=f"Configure console logging. Options: "
            f"{', '.join(k for k, v in sorted(LOG_LEVELS.items(), key=lambda x: x[1]))} (default: %(default)s)",
        )

        boundary_args = self.parser.add_argument_group(
            title="Boundary Arguments",
            description="Accepts revision or build date in YYYY-MM-DD format)",
        )
        boundary_args.add_argument(
            "--start",
            help="Start build id (default: earliest available build)",
        )
        boundary_args.add_argument(
            "--end",
            help="End build id (default: latest available build)",
        )

        bisection_args = self.parser.add_argument_group(title="Bisection Arguments")
        bisection_args.add_argument(
            "--timeout",
            type=int,
            default=60,
            help="Maximum iteration time in seconds (default: %(default)s)",
        )
        bisection_args.add_argument(
            "--repeat",
            type=int,
            default=1,
            help="Number of times to evaluate testcase (per build)",
        )
        bisection_args.add_argument(
            "--config",
            type=Path,
            help="Path to optional config file",
        )
        bisection_args.add_argument(
            "--find-fix",
            action="store_true",
            help="Identify fix date",
        )

        target_group = self.parser.add_argument_group("Target Arguments")
        target_group.add_argument(
            "--os",
            choices=sorted(Platform.SUPPORTED),
            help=f"Specify the target system. (default: {std_platform.system()})",
        )
        target_group.add_argument(
            "--cpu",
            choices=CPU_CHOICES,
            help=f"Specify the target CPU. (default: {std_platform.machine()})",
        )

        branch_group = self.parser.add_argument_group("Branch Arguments")
        branch_args = branch_group.add_mutually_exclusive_group()
        branch_args.add_argument(
            "--central",
            action="store_const",
            const="central",
            dest="branch",
            help="Download from mozilla-central (default)",
        )
        branch_args.add_argument(
            "--release",
            action="store_const",
            const="release",
            dest="branch",
            help="Download from mozilla-release",
        )
        branch_args.add_argument(
            "--beta",
            action="store_const",
            const="beta",
            dest="branch",
            help="Download from mozilla-beta",
        )
        branch_args.add_argument(
            "--esr-stable",
            action="store_const",
            const="esr-stable",
            dest="branch",
            help="Download from esr-stable",
        )
        branch_args.add_argument(
            "--esr-next",
            action="store_const",
            const="esr-next",
            dest="branch",
            help="Download from esr-next",
        )
        branch_args.add_argument(
            "--try",
            action="store_const",
            const="try",
            dest="branch",
            help="Download from try",
        )
        branch_args.add_argument(
            "--autoland",
            action="store_const",
            const="autoland",
            dest="branch",
            help="Download from autoland",
        )

        build_group = self.parser.add_argument_group("Build Arguments")
        build_group.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="Get debug builds w/ symbols (default=optimized).",
        )
        build_group.add_argument(
            "-a",
            "--asan",
            action="store_true",
            help="Download AddressSanitizer builds.",
        )
        build_group.add_argument(
            "-t",
            "--tsan",
            action="store_true",
            help="Download ThreadSanitizer builds.",
        )
        build_group.add_argument(
            "--fuzzing",
            action="store_true",
            help="Download --enable-fuzzing builds.",
        )
        build_group.add_argument(
            "--fuzzilli",
            action="store_true",
            help="Download --enable-js-fuzzilli builds.",
        )
        build_group.add_argument(
            "--coverage",
            action="store_true",
            help="Download --coverage builds.",
        )
        build_group.add_argument(
            "--valgrind",
            action="store_true",
            help="Download Valgrind builds.",
        )
        build_group.add_argument(
            "--no-opt",
            action="store_true",
            help="Download non-optimized builds.",
        )

        self.parser.set_defaults(branch="central")

    def sanity_check(self, args: Namespace) -> None:
        """
        Perform sanity checks.

        :param args: Parsed arguments.
        :raises SystemExit: If sanity check fails.
        """
        args.testcase = args.testcase.expanduser()
        if not args.testcase.is_file() or not os.access(args.testcase, os.R_OK):
            self.parser.error("Cannot access testcase!")

        log_level = LOG_LEVELS.get(args.log_level.upper(), None)
        if log_level is None:
            self.parser.error(f"Invalid log-level {args.log_level!r}")
        args.log_level = log_level

        if args.config is not None:
            args.config = args.config.expanduser()
            if not args.config.is_file() or not os.access(args.config, os.R_OK):
                self.parser.error("Cannot access configuration file!")

        if args.branch is None:
            args.branch = "central"
        elif args.branch.startswith("esr"):
            args.branch = Fetcher.resolve_esr(args.branch)

        if args.target == "firefox" and args.fuzzilli:
            self.parser.error("Fuzzilli builds are not available for firefox")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import argparse
import logging
import os
from pathlib import Path
from platform import system
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Optional, NoReturn, Any, List

from grizzly.common.frontend import Exit
from grizzly.common.storage import TestCase
from grizzly.replay import ReplayArgs, ReplayManager

from autobisect.evaluators.base import Evaluator, EvaluatorResult

LOG = logging.getLogger(__name__)


class ArgParserNoExit(argparse.ArgumentParser):
    """Override default ArgParser SystemExit Behavior"""

    def exit(self, status: int = 0, message: Optional[str] = None) -> NoReturn:
        raise SystemExit(f"Suppressed exit: status={status}, message={message}")

    def error(self, message: str) -> NoReturn:
        raise BrowserEvaluatorException(message)


class ReplayArgsNoExit(ReplayArgs):
    """Set parser to ArgParserNoExit instance"""

    def __init__(self) -> None:
        self.parser = ArgParserNoExit()
        super().__init__()


class BrowserEvaluatorException(Exception):
    """Simple exception handler for BrowserEvaluator."""


class BrowserEvaluator(Evaluator):
    """Testcase evaluator for Firefox."""

    target = "firefox"

    def __init__(self, testcase: Path, **kwargs: Any) -> None:
        self.testcase = testcase

        # Grizzly arguments
        self.env_vars = kwargs.get("env", None)
        self.display = kwargs.get("display", None)
        self.ignore = kwargs.get("ignore", None)
        self.launch_attempts = kwargs.get("launch_attempts", 15)
        self.launch_timeout = kwargs.get("launch_timeout", None)
        self.logs = kwargs.get("logs", None)
        self.pernosco = kwargs.get("pernosco", False)
        self.prefs = kwargs.get("prefs", None)
        self.relaunch = kwargs.get("relaunch", 1)
        self.repeat = kwargs.get("repeat", None)
        self.timeout = kwargs.get("timeout", None)
        self.time_limit = kwargs.get("time_limit", None)
        self.use_harness = kwargs.get("use_harness", None)
        self.valgrind = kwargs.get("valgrind", False)

        if logging.getLogger().level != logging.DEBUG:
            logging.getLogger("grizzly").setLevel(logging.INFO)

    def parse_args(
        self,
        binary: Path,
        test_dir: Path,
        verify: Optional[bool] = False,
    ) -> argparse.Namespace:
        """
        Parse arguments destined for grizzly.

        :param binary: The path to the firefox binary.
        :param test_dir: The path to the testcase.
        :param verify: Indicates if we're running a testcase or verifying the browser stability.
        :returns: The return code or None.
        """
        raw_args: List[Any] = [binary, test_dir]

        if self.launch_attempts:
            raw_args.extend(["--launch-attempts", self.launch_attempts])

        if self.ignore:
            raw_args.extend(["--ignore"] + self.ignore)
        if self.launch_timeout:
            raw_args.extend(["--launch-timeout", self.launch_timeout])
        if self.prefs:
            raw_args.extend(["--prefs", self.prefs])
        if self.relaunch:
            raw_args.extend(["--relaunch", self.relaunch])
        if self.timeout:
            raw_args.extend(["--timeout", self.timeout])
        if self.time_limit:
            raw_args.extend(["--time-limit", self.time_limit])
        if not self.use_harness:
            raw_args.append("--no-harness")
        if self.valgrind:
            raw_args.append("--valgrind")
        if self.display:
            raw_args.extend(["--display", self.display])

        if not verify:
            if self.logs is not None:
                raw_args.extend(["--output", self.logs])
            if self.pernosco:
                raw_args.extend(["--pernosco"])
            if self.repeat is not None:
                raw_args.extend(["--repeat", self.repeat])

        return ReplayArgsNoExit().parse_args([str(arg) for arg in raw_args])

    def verify_build(self, binary: Path) -> bool:
        """
        Verify that build doesn't crash on start.

        :param binary: The path to the target binary.
        :return: True if the build is valid, False otherwise.
        """
        with NamedTemporaryFile(suffix=".html", mode="w", delete=False) as temp:
            temp.write("<html><script>window.close()</script></html>")

        try:
            LOG.info("> Verifying build...")
            status = self.launch(binary, Path(temp.name), verify=True)

            if status != EvaluatorResult.BUILD_PASSED:
                LOG.error(">> Failed to validate build!")
                return False

            LOG.info(">> Build verified!")
            return True
        finally:
            os.unlink(temp.name)

    def evaluate_testcase(self, build_path: Path) -> EvaluatorResult:
        """
        Validate build and launch with supplied testcase.

        :returns: Result of evaluation.
        """
        binary = "firefox.exe" if system() == "Windows" else "firefox"
        binary_path = build_path / binary
        result = EvaluatorResult.BUILD_FAILED
        if not binary_path.is_file():
            LOG.error("Cannot find build path!")
            return result

        if self.verify_build(binary_path):
            LOG.info("> Launching build with testcase...")
            result = self.launch(binary_path, self.testcase, scan_dir=True)

            if result == EvaluatorResult.BUILD_CRASHED:
                LOG.info(">> Build crashed!")
            else:
                LOG.info(">> Build did not crash!")

        return result

    def launch(
        self,
        binary: Path,
        test_path: Path,
        verify: bool = False,
        scan_dir: bool = False,
    ) -> EvaluatorResult:
        """
        Launch firefox using the supplied binary and testcase.

        :param binary: The path to the firefox binary.
        :param test_path: The path to the testcase.
        :param verify: Indicates if we're running a testcase or verifying the browser stability.
        :param scan_dir: Scan subdirectory for additional files to serve.
        :returns: The return code or None.
        """
        if not binary.is_file():
            raise BrowserEvaluatorException(f"Binary path does not exist ({binary})!")

        # Create testcase
        testcase = TestCase.load(test_path, catalog=scan_dir)
        if self.env_vars:
            for key, value in self.env_vars.items():
                testcase.env_vars[key] = value

        with TemporaryDirectory() as test_dir:
            testcase.dump(Path(test_dir), include_details=True)
            args = self.parse_args(binary, Path(test_dir), verify)
            success = ReplayManager.main(args)

            if success == Exit.SUCCESS:
                return EvaluatorResult.BUILD_CRASHED
            if success == Exit.FAILURE:
                return EvaluatorResult.BUILD_PASSED

            return EvaluatorResult.BUILD_FAILED
